; Block save and load routines for ZX81 cassette tape.
; Modified from the Improved ROM. 
; AJB - January 2021

; There are 2 primary routines: Blk_Save & Blk_Load
; They take similar parameters:
;
; DE = Start address of block to save/load
; BC = Address of the last byte of the block to save/load
; HL = Address of filename. Filenames end with a terminator byte. This is
;      an extra byte with bit 7 set. Filenames have no quotes.
;      A terminator may be given on it's own with Blk_Load and it will cause
;      the next file on tape to be loaded.
;
; Both routines must be called from FAST mode.
; The source code below is written for the TASM assembler.
	
#define	BREAK-1	$0f46
#define	FAST	$0f23
#define SLOW	$0f2b

	.org	$4800		; Arbitrary address to test compilation.
				; 

Blk_S	.word	$00		; Start of block to Save/Load
Blk_E	.word	$00		; End of block to Save/Load

Fname	.byte	$26,$27,$28,$80	; ABC . Example filename




; BlkSave - Save a block of memory to cassette
; Enter: HL = Start address of name (terminated by a byte with bit 7 set)
;        DE = Start address of block to save
;	 BC = End address of block to save
;
; Exit: NC = Break pressed

Blk_Save	.MODULE BlkSave_M
	ld	(Blk_S),de
	ld	(Blk_E),bc

	ld	de,$12cb	; 5 seconds timing value
_1	call	BREAK-1		; Outer delay loop. No carry if SPACE
				; not pressed
	ret	nc

	ld	b,$00		; Inner delay loop
	djnz	*	
	dec	de		; decrement outer loop counter
	ld	a,d
	or	e
	jr	nz,_1		; and jump back if not zero.

	; Save name
_namout	call	Out_Byte	; Output character
	ret	nc		; Break pressed
	bit	7,(hl)		; Test for terminator
	inc	hl		; Increment name pointer
	jr	z,_namout

	; Save data
	ld	hl,(Blk_S)
_datout	call	Out_Byte
	ret	nc
	call	Blk_upd		; Test for end of block
	jr	nc,_datout
	ret




; Output character in (hl) to cassette
; Returns: NC=Break pressed

Out_Byte .MODULE Out_Byte_M
	ld	e,(hl)		; Get byte to save
	scf			; Set carry as a marker
_OBbit	rl	e		; Rotate bit into carry
	jr	nz,_dobit	; Return whan marker bit has passed through.
	scf			; Carry set denotes success.
	ret			
_dobit	sbc	a,a		; Set bit=$ff Clear bit=$00
	and	$05		; Set bit=$05 Clear bit=$00
	add	a,$04		; Set bit=$09 Clear bit=$04
	ld	c,a		; Move timer value to C

_OBpulse
	out	($ff),a		; Pulse and delay
	ld	b,$23
	djnz	*
	call	BREAK-1		; Test for SPACE
	ret	nc		; Break

	ld	b,$1e		; Delay
	djnz	*

	dec	c		; Decrement pulse number count
	jr	nz,_OBpulse	; Loop back if more pulses required

	and	a		; Clear carry for next bit test
	djnz	*-1
	jr	_OBbit		; Loop back for next bit




; Enter: HL = Start address of name (terminated by a byte with bit 7 set)
;        DE = Start address of block to save
;	 BC = End address of block to save
;
; Exit: NC = Break pressed

Blk_Load
	ex	de,hl
	ld	(Blk_S),hl
	ld	(Blk_E),bc

_BLnxtp	call	_InByte
	ret	nc
	jr	_BLnxtp




; Get a byte from cassette
_InByte
	ld	c,%00000001	; Prepare 8-bit counter
_IBnxtb	ld	b,$00		; Counter set to 256

_IBtlp	ld	a,$7f		; Read kdb row with
	in	a,($fe)		; the SPACE key
	out	($ff),a		; Output signal to screen
	rra			; Test for key pressed
	ret	nc

	rla			; Reverse last rotation then
	rla			; rotate to get tape test bit
	jr	c,_IBgetbit
	djnz	_IBtlp

	pop	af		; Timeout test
	ld	a,$03		; Force MSB of return address to imitate ROM
	cp	d
	ret	nc

	ld	h,d		; Transfer name start address to HL
	ld	l,e

_InName
	call	_InByte		; Get byte
	bit	7,d		; Is name the null string?
	ld	a,c		; Transfer byte to A
	jr	nz,_IBmatch	; Forward with null string to _IBmatch

	cp	(hl)		; else compare with string in memory
	jr	nz,_BLnxtp	; Go back on mismatch


_IBmatch
	inc	hl		; Address of next character
	rla			; Test for inverted bit
	jr	nc,_InName	; Back if not.

	ld	hl,(Blk_S)
_InProg
	ld	d,b		; Set D to 0 as indicator
	call	_InByte
	ld	(hl),c		; Insert assembled byte
	call	Blk_upd
	jr	nc,_InProg
	ret


_IBgetbit
	push	de		; Save the timing value
	ld	e,$94
_IBtrail
	ld	b,$1a
_IBcnt	dec	e		; Decrement measuring timer
	in	a,($fe)		; Read tape
	rla
	bit	7,e
	ld	a,e
	jr	c,_IBtrail
	djnz	_IBcnt

	pop	de
	jr	nz,_IBbdone
	cp	$56
	jr	nc,_IBnxtb
_IBbdone
	ccf
	rl	c
	jr	nc,_IBnxtb
	ret




; Test for data block end.
; Enter: HL = address of last output character.
; Exit: NC = more bytes to save

Blk_upd
	inc	hl
	ex	de,hl
	ld	hl,(Blk_E)
	and	a
	sbc	hl,de
	ex	de,hl
	ret

	.end
